@@ -1,6 +1,6 @@
 /* Delete entries from a tar archive.
 
-   Copyright 1988-2021 Free Software Foundation, Inc.
+   Copyright 1988-2023 Free Software Foundation, Inc.
 
    This file is part of GNU tar.
 
@@ -18,7 +18,6 @@
    along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
 
 #include <system.h>
-#include <system-ioctl.h>
 
 #include "common.h"
 #include <rmt.h>
@@ -50,41 +49,31 @@ move_archive (off_t count)
   if (count == 0)
     return;
 
-#ifdef MTIOCTOP
-  {
-    struct mtop operation;
-
-    if (count < 0
-	? (operation.mt_op = MTBSR,
-	   operation.mt_count = -count,
-	   operation.mt_count == -count)
-	: (operation.mt_op = MTFSR,
-	   operation.mt_count = count,
-	   operation.mt_count == count))
-      {
-	if (0 <= rmtioctl (archive, MTIOCTOP, (char *) &operation))
-	  return;
-
-	if (errno == EIO
-	    && 0 <= rmtioctl (archive, MTIOCTOP, (char *) &operation))
-	  return;
-      }
-  }
-#endif /* MTIOCTOP */
-
-  {
-    off_t position0 = rmtlseek (archive, (off_t) 0, SEEK_CUR);
-    off_t increment = record_size * (off_t) count;
-    off_t position = position0 + increment;
-
-    if (increment / count != record_size
-	|| (position < position0) != (increment < 0)
-	|| (position = position < 0 ? 0 : position,
-	    rmtlseek (archive, position, SEEK_SET) != position))
-      seek_error_details (archive_name_array[0], position);
-
+  if (mtioseek (false, count))
     return;
-  }
+
+  off_t position0 = rmtlseek (archive, 0, SEEK_CUR), position = 0;
+  if (0 <= position0)
+    {
+      /* Pretend the starting position is at the first record
+	 boundary after POSITION0.  This is useful at EOF after
+	 a short read.  */
+      idx_t short_size = position0 % record_size;
+      idx_t start_offset = short_size ? record_size - short_size : 0;
+      off_t increment, move_start;
+      if (INT_MULTIPLY_WRAPV (record_size, count, &increment)
+	  || INT_ADD_WRAPV (position0, start_offset, &move_start)
+	  || INT_ADD_WRAPV (move_start, increment, &position)
+	  || position < 0)
+	{
+	  ERROR ((0, EOVERFLOW, "lseek: %s", archive_name_array[0]));
+	  return;
+	}
+      else if (rmtlseek (archive, position, SEEK_SET) == position)
+	return;
+    }
+  if (!_isrmt (archive))
+    seek_error_details (archive_name_array[0], position);
 }
 
 /* Write out the record which has been filled.  If MOVE_BACK_FLAG,
@@ -147,6 +136,26 @@ write_recent_bytes (char *data, size_t bytes)
     write_record (1);
 }
 
+static void
+flush_file (void)
+{
+  off_t blocks_to_skip;
+
+  set_next_block_after (current_header);
+  blocks_to_skip = (current_stat_info.stat.st_size
+			  + BLOCKSIZE - 1) / BLOCKSIZE;
+
+  while (record_end - current_block <= blocks_to_skip)
+    {
+      blocks_to_skip -= (record_end - current_block);
+      flush_archive ();
+      if (record_end == current_block)
+	/* Hit EOF */
+	break;
+    }
+  current_block += blocks_to_skip;
+}
+
 void
 delete_archive_members (void)
 {
@@ -155,7 +164,6 @@ delete_archive_members (void)
 
   /* FIXME: Should clean the routine before cleaning these variables :-( */
   struct name *name;
-  off_t blocks_to_skip = 0;
   off_t blocks_to_keep = 0;
   int kept_blocks_in_record;
 
@@ -163,11 +171,12 @@ delete_archive_members (void)
   open_archive (ACCESS_UPDATE);
   acting_as_filter = strcmp (archive_name_array[0], "-") == 0;
 
+  /* Skip to the first member that matches the name list. */
   do
     {
       enum read_header status = read_header (&current_header,
-                                             &current_stat_info,
-                                             read_header_x_raw);
+					     &current_stat_info,
+					     read_header_x_raw);
 
       switch (status)
 	{
@@ -175,15 +184,15 @@ delete_archive_members (void)
 	  abort ();
 
 	case HEADER_SUCCESS:
-	  if ((name = name_scan (current_stat_info.file_name)) == NULL)
+	  if ((name = name_scan (current_stat_info.file_name, false)) == NULL)
 	    {
-	      skip_member ();
+	      skim_member (acting_as_filter);
 	      break;
 	    }
 	  name->found_count++;
-	  if (!ISFOUND(name))
+	  if (!ISFOUND (name))
 	    {
-	      skip_member ();
+	      skim_member (acting_as_filter);
 	      break;
 	    }
 	  FALLTHROUGH;
@@ -243,15 +252,12 @@ delete_archive_members (void)
 
       if (logical_status == HEADER_SUCCESS)
 	{
-	  /* FIXME: Pheew!  This is crufty code!  */
 	  logical_status = HEADER_STILL_UNREAD;
-	  goto flush_file;
+	  flush_file ();
 	}
 
-      /* FIXME: Solaris 2.4 Sun cc (the ANSI one, not the old K&R) says:
-	 "delete.c", line 223: warning: loop not entered at top
-	 Reported by Bruno Haible.  */
-      while (1)
+      /* Skip matching members and move the rest up the archive. */
+      while (logical_status != HEADER_END_OF_FILE)
 	{
 	  enum read_header status;
 
@@ -259,105 +265,108 @@ delete_archive_members (void)
 
 	  if (current_block == record_end)
 	    flush_archive ();
+
 	  status = read_header (&current_header, &current_stat_info,
-	                        read_header_auto);
+				read_header_auto);
 
-	  xheader_decode (&current_stat_info);
-
-	  if (status == HEADER_ZERO_BLOCK && ignore_zeros_option)
+	  switch (status)
 	    {
-	      set_next_block_after (current_header);
-	      continue;
-	    }
-	  if (status == HEADER_END_OF_FILE || status == HEADER_ZERO_BLOCK)
-	    {
-	      logical_status = HEADER_END_OF_FILE;
-	      break;
-	    }
+	    case HEADER_STILL_UNREAD:
+	    case HEADER_SUCCESS_EXTENDED:
+	      abort ();
 
-	  if (status == HEADER_FAILURE)
-	    {
-	      ERROR ((0, 0, _("Deleting non-header from archive")));
-	      set_next_block_after (current_header);
-	      continue;
-	    }
+	    case HEADER_SUCCESS:
+	      /* Found another header.  */
+	      xheader_decode (&current_stat_info);
 
-	  /* Found another header.  */
-
-	  if ((name = name_scan (current_stat_info.file_name)) != NULL)
-	    {
-	      name->found_count++;
-	      if (ISFOUND(name))
+	      if ((name = name_scan (current_stat_info.file_name, false)) != NULL)
 		{
-		flush_file:
-		  set_next_block_after (current_header);
-		  blocks_to_skip = (current_stat_info.stat.st_size
-				    + BLOCKSIZE - 1) / BLOCKSIZE;
-
-		  while (record_end - current_block <= blocks_to_skip)
+		  name->found_count++;
+		  if (ISFOUND (name))
 		    {
-		      blocks_to_skip -= (record_end - current_block);
-		      flush_archive ();
+		      flush_file ();
+		      break;
 		    }
-		  current_block += blocks_to_skip;
-		  blocks_to_skip = 0;
-		  continue;
 		}
-	    }
-	  /* Copy header.  */
+	      /* Copy header.  */
 
-	  if (current_stat_info.xhdr.size)
-	    {
-	      write_recent_bytes (current_stat_info.xhdr.buffer,
-				  current_stat_info.xhdr.size);
-	    }
-	  else
-	    {
-	      write_recent_blocks (recent_long_name, recent_long_name_blocks);
-	      write_recent_blocks (recent_long_link, recent_long_link_blocks);
-	    }
-	  new_record[new_blocks] = *current_header;
-	  new_blocks++;
-	  blocks_to_keep
-	    = (current_stat_info.stat.st_size + BLOCKSIZE - 1) / BLOCKSIZE;
-	  set_next_block_after (current_header);
-	  if (new_blocks == blocking_factor)
-	    write_record (1);
-
-	  /* Copy data.  */
-
-	  kept_blocks_in_record = record_end - current_block;
-	  if (kept_blocks_in_record > blocks_to_keep)
-	    kept_blocks_in_record = blocks_to_keep;
-
-	  while (blocks_to_keep)
-	    {
-	      int count;
-
-	      if (current_block == record_end)
+	      if (current_stat_info.xhdr.size)
 		{
-		  flush_read ();
-		  current_block = record_start;
-		  kept_blocks_in_record = blocking_factor;
-		  if (kept_blocks_in_record > blocks_to_keep)
-		    kept_blocks_in_record = blocks_to_keep;
+		  write_recent_bytes (current_stat_info.xhdr.buffer,
+				      current_stat_info.xhdr.size);
 		}
-	      count = kept_blocks_in_record;
-	      if (blocking_factor - new_blocks < count)
-		count = blocking_factor - new_blocks;
-
-	      if (! count)
-		abort ();
-
-	      memcpy (new_record + new_blocks, current_block, count * BLOCKSIZE);
-	      new_blocks += count;
-	      current_block += count;
-	      blocks_to_keep -= count;
-	      kept_blocks_in_record -= count;
-
+	      else
+		{
+		  write_recent_blocks (recent_long_name,
+				       recent_long_name_blocks);
+		  write_recent_blocks (recent_long_link,
+				       recent_long_link_blocks);
+		}
+	      new_record[new_blocks] = *current_header;
+	      new_blocks++;
+	      blocks_to_keep
+		= (current_stat_info.stat.st_size + BLOCKSIZE - 1) / BLOCKSIZE;
+	      set_next_block_after (current_header);
 	      if (new_blocks == blocking_factor)
 		write_record (1);
+
+	      /* Copy data.  */
+
+	      kept_blocks_in_record = record_end - current_block;
+	      if (kept_blocks_in_record > blocks_to_keep)
+		kept_blocks_in_record = blocks_to_keep;
+
+	      while (blocks_to_keep)
+		{
+		  int count;
+
+		  if (current_block == record_end)
+		    {
+		      flush_read ();
+		      current_block = record_start;
+		      kept_blocks_in_record = blocking_factor;
+		      if (kept_blocks_in_record > blocks_to_keep)
+			kept_blocks_in_record = blocks_to_keep;
+		    }
+		  count = kept_blocks_in_record;
+		  if (blocking_factor - new_blocks < count)
+		    count = blocking_factor - new_blocks;
+
+		  if (! count)
+		    abort ();
+
+		  memcpy (new_record + new_blocks, current_block,
+			  count * BLOCKSIZE);
+		  new_blocks += count;
+		  current_block += count;
+		  blocks_to_keep -= count;
+		  kept_blocks_in_record -= count;
+
+		  if (new_blocks == blocking_factor)
+		    write_record (1);
+		}
+	      break;
+
+	    case HEADER_ZERO_BLOCK:
+	      if (ignore_zeros_option)
+		set_next_block_after (current_header);
+	      else
+		logical_status = HEADER_END_OF_FILE;
+	      break;
+
+	    case HEADER_END_OF_FILE:
+	      logical_status = HEADER_END_OF_FILE;
+	      break;
+
+	    case HEADER_FAILURE:
+	      ERROR ((0, 0, _("Deleting non-header from archive")));
+	      set_next_block_after (current_header);
+	      break;
+
+	    default:
+	      abort ();
 	    }
+	  tar_stat_destroy (&current_stat_info);
 	}
 
       if (logical_status == HEADER_END_OF_FILE)
